/* --------------------------------------------------------------------------
 * Hit It! 
 * 
 * A kinect hack
 * by @ThisThatLab (http://thisthatlab.com) and LoVid (http://lovid.org)
 *
 * Built with the simple-openni Processing Wrapper for the OpenNI/Kinect library
 *
 * To play the game you must first install the OpenNI/Kinect library. Go here and follow the instructions.
 * http://code.google.com/p/simple-openni
 * ----------------------------------------------------------------------------
 *
 * This game plays best with multiple hands. To enable multiple hands:
 *
 * on Mac, change the .ini file at this location: /usr/etc/primesense/XnVHandGenerator/Nite.ini to look like this
 *  [HandTrackerManager]
 *  AllowMultipleHands=1
 *  TrackAdditionalHands=1
 * ( to be safe, also make these change here:  /usr/etc/primesense/Hands_1_5_2/Nite.ini)
 *
 * on Windows you can find the file at:
 *  C:\Program Files (x86)\Prime Sense\NITE\Hands\Data\Nite.ini
 * ----------------------------------------------------------------------------
 */

import SimpleOpenNI.*;
import ddf.minim.*;

SimpleOpenNI      context;

// NITE //
XnVSessionManager sessionManager;
XnVFlowRouter     flowRouter;
boolean showDepthImage = false;

PointDrawer       pointDrawer;

Minim minim;
AudioSample[] hitSounds;
AudioPlayer themeSound;

color b1, b2, b3;
int hours, minutes, seconds, initMillis;
String timeString;
PFont font, fontLarge;
color newGameCol;
int newGameX, newGameY;
int gameOverTimer, gameOverTimerThresh;
int gameTime;

Sprite[] sprites = new Sprite[5];
int[] spriteTypes = {0, 1, 2, 3, 4};

PImage[][] allSprites;
String[] typeName = { "boombox", "monkey", "pineapple", "pirate", "robot" };
String[] colorName = { "WHITE", "RED", "GREEN", "BLUE" };
String[] audioFiles = { "boombox.wav", "monkey.wav", "pineapple.wav", "pirate.wav", "robot.wav" };
int numStates = 4;

boolean hittable = false;
int hittableDuration;

int hitTime;
int hittableTime;

boolean gameOver, gameStart;
int gameOverRad = 300;

void setup() {
 	context = new SimpleOpenNI(this);
  
 	// mirror is by default enabled
 	context.setMirror(true);
 
 	// enable depthMap generation 
 	if(context.enableDepth() == false ) {
   	println("Can't open the depthMap, maybe the camera is not connected!"); 
    exit();
    return;
 	}
 
 	// enable the hands + gesture
 	context.enableGesture();
 	context.enableHands();

 	// setup NITE 
 	sessionManager = context.createSessionManager("Click,Wave", "RaiseHand");

 	pointDrawer = new PointDrawer();
 	flowRouter = new XnVFlowRouter();
 	flowRouter.SetActive(pointDrawer);
 
 	sessionManager.AddListener(flowRouter);
          
 	size(context.depthWidth(), context.depthHeight()); 
 	smooth();

	allSprites = new PImage[5][4];
	gameOver = false;
	gameStart = true;
	
	// background
	b1 = color(100, 0, 100);
	b2 = color(20, 20, 20);
	b3 = color(180, 180, 0);
	
	newGameCol = color(255,0,0);
	newGameX = 40;
	newGameY = 40;
	gameOverTimer = 0;
	gameOverTimerThresh = 100;
	
	font = loadFont("Courier-32.vlw");
	fontLarge = loadFont("Courier-64.vlw"); 
	
	// load sounds
	minim = new Minim(this);
	hitSounds = new AudioSample[5];
	for (int i=0; i<5; i++) {
		hitSounds[i] = minim.loadSample("audio/" + audioFiles[i], 2048);
	}	
	themeSound = minim.loadFile("audio/theme.wav", 2048);
	themeSound.loop();
	
	// init game
	initMillis = millis();
	initSprites();
	newHittable();
}

void draw() {
	setGradient(0, 0, width, height, b1, b2);
	
	// update the cam
  context.update();

  // update nite
  context.update(sessionManager);

  // display kinect's depthImageMap by typing 'q'
  if (showDepthImage) {
  	image(context.depthImage(),0,0);
	}
	
	if (hittable && millis()-hittableTime > hittableDuration) {
		newHittable();
	}
	
	if (gameStart) {
		fill(b3);
		textAlign(CENTER);
		textFont(fontLarge);
		text("HIT IT!", width/2, height/2);
		textFont(font);
		text("DIRECTIONS:", width/2, height/2+100);
		textFont(font, 16);
		text("Wave your hands from side to side\nuntil you get a red and a green trail.\nThen hit the red spot with your red hand to play.", width/2, height/2+130);
	
    noStroke();
	  fill(newGameCol);
	  ellipse(newGameX, newGameY, 50, 50);
	} else if (gameOver) {
		fill(b3);
		textAlign(CENTER);
		textFont(font);
		text("GREAT JOB!!! YOU WIN!!!", width/2, height/2);
                
	  textAlign(RIGHT);
	  text("YOUR TIME:", width/2+40, height/2+50);
	  textAlign(LEFT);
	  drawTimer(gameTime, width/2+50, height/2+50);
	
		if (gameOverTimer < gameOverTimerThresh) {
			gameOverTimer++;
		} else {
			noStroke();
			fill(newGameCol);
			ellipse(newGameX, newGameY, 50, 50);
			textFont(font, 16);
			textAlign(LEFT);
			text("Hit the red spot to play again.", 82, 42);
		}
		
	} else {
		for (int i=0; i<3; i++) {
			addBGLine();
		}

		// draw each of the sprites
	  for (int i=0; i<sprites.length; i++) { 
	  	sprites[i].update();
	  }

		// draw timer
    textAlign(LEFT);
    fill(255);
	  textFont(font);
    seconds = floor((millis()-initMillis)/1000);
		drawTimer(seconds, 10, height-10);
	}
	
	// draw the hand trails
 	pointDrawer.draw();
}

void addBGLine() {
	float dir = random(0,1);
	stroke(b3);
	strokeWeight(random(1,3));
	if (dir < 0.5) {
    	line(random(width), 0, random(width), height);
  	} else {
		line(0, random(height), width, random(height));
	}
}

void setGradient(int x, int y, float w, float h, color c1, color c2){
  	// calculate differences between color components 
  	float deltaR = red(c2)-red(c1);
  	float deltaG = green(c2)-green(c1);
  	float deltaB = blue(c2)-blue(c1);

    for (int i=x; i<=(x+w); i++){
    	// row
      	for (int j = y; j<=(y+h); j++){
        	color c = color(
        		(red(c1)+(j-y)*(deltaR/h)),
        		(green(c1)+(j-y)*(deltaG/h)),
        		(blue(c1)+(j-y)*(deltaB/h)) 
          	);
        	set(i, j, c);
      	}
    }  
}

void drawTimer(int useSeconds, int timeX, int timeY) {
	
	hours = floor(useSeconds/3600);
	useSeconds -= hours*3600;
	minutes = floor(useSeconds/60);
	useSeconds -= minutes*60;
	
	timeString = "";
	boolean padMin, padSec;
	if (hours > 0) {
		timeString += hours + ":";
		padMin = true;
	} else {
		padMin = false;
	}
	
	if (padMin) {
		if (minutes == 0) {
			timeString += "00:";
		} else if (minutes < 10) {
			timeString += "0" + minutes + ":";
		} else {
			timeString += minutes + ":";
		}
	} else if (minutes > 0) {
		timeString += minutes + ":";
	}
	
	if (minutes > 0) {
		padSec = true;
	} else {
		padSec = false;
	}
	
	if (padSec && useSeconds < 10) {
		timeString += "0" + useSeconds;
	} else {
		timeString += useSeconds;
	}

	text(""+ timeString, timeX, timeY);
}

void keyPressed() {
	switch(key) {
  	case 'e':
   		// end sessions
   		sessionManager.EndSession();
   		println("end session");
   		break;
		case 'q':
			showDepthImage = !showDepthImage;
			break;
  }
}


////////////////////////////////////////
// Sprite related code

void initSprites() {
	// load sprite images
	String path;
	for (int i=0; i<5; i++) {
		for (int j=0; j<numStates; j++) {
			path = "img/"+ typeName[i] +"/"+ typeName[i] + colorName[j] + ".png";
			println("Loading: " + path);
			allSprites[i][j] = loadImage(path);
		}
	} 
	
	// create sprites
	for (int i=0; i<sprites.length; i++) {
    	sprites[i] = new Sprite( i, spriteTypes[i] );
	}  
}

void newHittable() {
	
	// get current hittable
  int curr = -1;
  for (int i=0; i<sprites.length; i++) {
    if ( sprites[i].hitState != Sprite.UNHITTABLE) {
    	curr = i;
      break;
    }
  }  
	if (curr != -1) 
		sprites[curr].makeUnhittable();
  
	// choose new hittable
  int index = floor(random(0,sprites.length));
  while (index == curr) {
		index = floor(random(0,sprites.length));
	}
	sprites[index].makeHittable();
}

void checkHit(float xPos, float yPos, color col) {
 	for (int i=0; i<sprites.length; i++) {
   	if ( sprites[i].hitState != Sprite.UNHITTABLE && col == sprites[i].col) {
     		sprites[i].checkHit(xPos, yPos);
   	}
 	}
}

void checkPlayAgain(float xPos, float yPos, color col) {
	if (col == newGameCol) {
		if (xPos >= newGameX-25 && xPos <= newGameX+25 && yPos >= newGameY-25 && yPos <= newGameY+25) {
			resetGame();
		}
	}
}

void resetGame() {
	for (int i=0; i<sprites.length; i++) {
  	sprites[i].reset();
  }
  if (gameStart) gameStart = false;
  gameTime = 0;
  initMillis = millis();
	gameOver = false;
	gameOverTimer = 0;
	newHittable();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// session callbacks

void onStartSession(PVector pos)
{
  println("onStartSession: " + pos);
}

void onEndSession()
{
  println("onEndSession: ");
}

void onFocusSession(String strFocus,PVector pos,float progress)
{
  println("onFocusSession: focus=" + strFocus + ",pos=" + pos + ",progress=" + progress);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
// PointDrawer keeps track of the handpoints

class PointDrawer extends XnVPointControl
{
  HashMap    _pointLists;
  int        _maxPoints;
  color[]    _colorList = { color(255,0,0),color(0,255,0),color(0,0,255),color(255,255,0)};
  
  public PointDrawer()
  {
    _maxPoints = 20;
    _pointLists = new HashMap();
  }
	
  public void OnPointCreate(XnVHandPointContext cxt)
  {
    // create a new list
    addPoint(cxt.getNID(),new PVector(cxt.getPtPosition().getX(),cxt.getPtPosition().getY(),cxt.getPtPosition().getZ()));
    
    //println("OnPointCreate, handId: " + cxt.getNID());
  }
  
  public void OnPointUpdate(XnVHandPointContext cxt)
  {
    //println("OnPointUpdate " + cxt.getPtPosition());   
    addPoint(cxt.getNID(),new PVector(cxt.getPtPosition().getX(),cxt.getPtPosition().getY(),cxt.getPtPosition().getZ()));
  }
  
  public void OnPointDestroy(long nID)
  {
    //println("OnPointDestroy, handId: " + nID);
    
    // remove list
    if(_pointLists.containsKey(nID))
       _pointLists.remove(nID);
  }
  
  public ArrayList getPointList(long handId)
  {
    ArrayList curList;
    if(_pointLists.containsKey(handId))
      curList = (ArrayList)_pointLists.get(handId);
    else
    {
      curList = new ArrayList(_maxPoints);
      _pointLists.put(handId,curList);
    }
    return curList;  
  }
  
  public void addPoint(long handId,PVector handPoint)
  {
    ArrayList curList = getPointList(handId);
    
    curList.add(0,handPoint);      
    if(curList.size() > _maxPoints)
      curList.remove(curList.size() - 1);
  }
  
  public void draw()
  {
    if(_pointLists.size() <= 0)
      return;
      
    pushStyle();
      noFill();
      
      PVector vec;
      PVector firstVec;
      PVector screenPos = new PVector();
      int colorIndex=0;
      
      // draw the hand lists
      Iterator<Map.Entry> itrList = _pointLists.entrySet().iterator();
      while(itrList.hasNext()) 
      {
        strokeWeight(2);
        color thisCol = _colorList[colorIndex % (_colorList.length - 1)];
        stroke(thisCol);

        ArrayList curList = (ArrayList)itrList.next().getValue();     
        
        // draw line
        firstVec = null;
        Iterator<PVector> itr = curList.iterator();
        beginShape();
          while (itr.hasNext()) 
          {
            vec = itr.next();
            if(firstVec == null)
              firstVec = vec;
            // calc the screen pos
            context.convertRealWorldToProjective(vec,screenPos);
            vertex(screenPos.x,screenPos.y);    
          } 
        endShape();   
  
        // draw current pos of the hand
        if(firstVec != null)
        {
          strokeWeight(8);
          context.convertRealWorldToProjective(firstVec,screenPos);
          point(screenPos.x,screenPos.y);
          
          // check coords
		  if (gameOver || gameStart) {
		  	checkPlayAgain(screenPos.x, screenPos.y, thisCol);
		  } else {
			checkHit(screenPos.x, screenPos.y, thisCol);
		  }
        }
        colorIndex++;
      }
      
    popStyle();
  }

}
