class Sprite {
	// types
	public static final int BOOMBOX = 0;
	public static final int MONKEY = 1;
	public static final int PINEAPPLE = 2;
	public static final int PIRATE = 3;
	public static final int ROBOT = 4;
	
	// hit states
	public static final int UNHITTABLE = 0;
	public static final int HITTABLE_RED = 1;
	public static final int HITTABLE_GREEN = 2;
	public static final int HIT_BLUE = 3;
	
	// miscellaneous constants
	public static final int radInit = 25;
	public static final int radStep = 25;
	public static final float gravity = 1.02;
	public static final int robotStep = 5;
	
	int id; 
  int typeID;
	
	float x, y;						// position
  float speedX, speedY;	// velocity
  int rad;							// radius
   	
	boolean visible;
  int hitState;
  color col;
   
	int boomboxDuration = 1500;
	int generalDuration = 3000;

	PImage icons[];

  // sprite constructor
  Sprite(int _id, int _typeID) {
	
		id = _id;
		typeID = _typeID;
		
		// assign icons
		icons = new PImage[4];
		for (int i=0; i<numStates; i++) {
			icons[i] = allSprites[typeID][i];
		}
		
		// init state
		reset();
	}
    
  void reset() {
		rad = radInit;
		hitState = UNHITTABLE;
		
		// set position
		setNewPosition(rad);
      	
		// set velocity
		switch(typeID) {
			case BOOMBOX:
				visible = false;
				speedX = speedY = 0;
				break;
			case MONKEY:
			case PIRATE:
				visible = true;
				speedX = random(0.75,1.5);
				speedY = random(0.75,1.5);
				distributeSpeed();
				break;
			case PINEAPPLE:
				visible = true;
				speedX = 0;
				speedY = 1;
				break;
			case ROBOT:
				visible = true;
				speedX = speedY = 0;
				break;
		}
	}
	
  void update() {
		updatePosition();
		updateState();
		
		if (visible) {
    	draw();
    }
  }
	
	void draw() {
		image(icons[hitState], x-rad, y-rad, rad*2, rad*2);
  }
	
  void updatePosition() {
		switch(typeID) {
			case BOOMBOX:
				break;
			case MONKEY:
			case PIRATE:
				x += speedX;
		    y += speedY;
				break;
			case PINEAPPLE:
				y += speedY;
				speedY *= gravity;
				break;
			case ROBOT:
				x += random(-robotStep, robotStep);
				y += random(-robotStep, robotStep);
				break;
		}
  }

	void updateState() {
		// check behaviors
		switch(typeID) {
			case BOOMBOX:
				break;
			case MONKEY:
				checkBounce();
				break;
			case PINEAPPLE:
				checkDrop();
				break;
			case PIRATE:
				checkWrap();
				break;
			case ROBOT:
				checkWander();
				break;
		}
	}	
   	
	// helper function to randomize position
	void setNewPosition(int padding) {
		x = random(padding*2, width-(padding*2));
    y = random(padding*2, height-(padding*2));
	}
	
	// helper function to randomize velocity
	void distributeSpeed() {
		float dir = random(0,1);
      	if (dir < 0.5) {
        	speedX *= -1;
      	}
      	dir = random(0,1);
      	if (dir < 0.5) {
        	speedY *= -1;
      	}
	}

	void makeUnhittable() {
		if (typeID == BOOMBOX) {
			visible = false;
      setNewPosition(rad);
		}
		
		hitState = UNHITTABLE;
    col = color(255); 
  }
   	
	void makeHittable() {
		if (typeID == BOOMBOX) {
			visible = true;
			hittableDuration = boomboxDuration;
		} else {
			hittableDuration = generalDuration;
		}
		
		// randomly choose which hand must hit it
    float c = random(0,1);
    if (c >= 0.5) {
			hitState = HITTABLE_RED;
      col = color(255,0,0);
    } else {
			hitState = HITTABLE_GREEN;
      col = color(0,255,0);
    }

		hittable = true;
		hittableTime = millis();
  }

	void makeHit() {
		hitState = HIT_BLUE;
   	col = color(0,0,255);
		hitSounds[typeID].trigger();
		hitTime = millis();
	}

 	void checkHit(float xPos, float yPos) {
  	if (xPos >= x-rad && xPos <= x+rad && yPos >= y-rad && yPos <= y+rad) {
			// succcessful hit!
    	rad += radStep;
			speedX *= 1.025;
			speedY *= 1.025;
			generalDuration -= 50;
			makeHit();
			println("hit! new radius:" + rad);
			if (rad > gameOverRad) {
				gameOver = true;
        gameTime = seconds;
			}
  	}
 	} 
   
	/*
	 * In this behavior when an object hits an edge it bounces in the opposite direction
	 */
	void checkBounce() {
		if (x < rad) x = rad;
		else if (x > width-rad) x = width-rad;
		
		if (x == width-rad || x == rad) {
    	speedX *= -1;
    }

		if (y < rad) y = rad;
		else if (y > height-rad) y = height-rad;
		
		if (y == height-rad || y == rad) {
      speedY *= -1;
    }
	}
	
	/*
	 * In this behavior when an object hits an edge it wraps to the opposite edge
	 */
	void checkWrap() {
		if (x >= width+rad) {
    	x = -rad;
    } else if (x <= -rad) {
			x = width+rad;
		}
		if (y >= height+rad) {
    	y = -rad;
    } else if (y <= -rad) {
			y = height+rad;
		}
	}
	
	/*
	 * In this behavior if an object wanders off the screen it returns to the center of the screen
	 */
	void checkWander() {
		if (x >= width || x <= 0 || y >= height || y <= 0) {
    	x = width/2;
			y = height/2;
    }
	}
	
	/*
	 * In this behavior if an object drops past the bottom of the screen it resets to a new random position
	 */
	void checkDrop() {
		if (y >= height+rad) {
    	setNewPosition(rad);
			speedY = 1;
    }
	}

}
